/*
 * CGI email client - main routines
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include "library.h"
#include "cgi.h"
#include "hash.h"
#include "smalloc.h"
#include "vector.h"
#include "log.h"
#include "rope.h"

#define SENDMAIL	"/usr/lib/sendmail -oi -oem -t"	/*No f!!king clue*/
#define PGP			"/usr/local/bin/pgp"
#define PGPDIR		"/usr/local/lib/pgp"
#define MULTIPART	"------------0123456789ABCDEF++"
#define SERIALSTR	"142857"	/* XXX Fill this in */

void parser_parse(hash_t *cgihash, rope_t *rope, FILE *output);
void config_run(hash_t *cgihash, char *hostname, char *template,
				rope_t *mrope);

/* This function bounces from PGP to sendmail, if we use PGP */
static void
sendmail_bounce(int pgpfd, FILE *sendmail)
{
	char	 line[BUFSIZ];
	FILE	*pgpout;

	pgpout = fdopen(pgpfd, "r");
	if (!pgpout) {
		log("cgisys", PRI_CRIT, "Could not fdopen PGP output "
						"filedes: %s", strerror(errno));
		exit(0);
	}

	while (fgets(line, BUFSIZ, pgpout)) {
		fprintf(sendmail, "%s", line);
	}

	fprintf(sendmail, "\n");
	fprintf(sendmail, "--" MULTIPART "--\n");

	fclose(pgpout);
	fclose(sendmail);
}

static char *
serialname(void)
{
	char	 line[BUFSIZ];
	snprintf(line, BUFSIZ, "%d%d", time(NULL), getpid());
	return sstrdup(line);
}

static void
mail_header(hash_t *cgihash, FILE *sendmail)
{
	char	*temp;

	fprintf(sendmail, "To: %s\n", (char *)hash_get(cgihash, "To"));
	fprintf(sendmail, "From: %s\n", (char *)hash_get(cgihash, "From"));
	fprintf(sendmail, "Subject: %s\n",
					(char *)hash_get(cgihash, "Subject"));

	temp = hash_get(cgihash, "Reply_to");
	if (temp && strlen(temp))
		fprintf(sendmail, "Reply-To: %s\n", temp);

	temp = hash_get(cgihash, "Cc");
	if (temp && strlen(temp))
		fprintf(sendmail, "Cc: %s\n", temp);
}

static void
pgp_header(hash_t *cgihash, FILE *sendmail)
{
	fprintf(sendmail, "Mime-Version: 1.0\n");
	fprintf(sendmail, "Content-Type: multipart/mixed; boundary="
					MULTIPART "\n");
	fprintf(sendmail, "\n");

/* Multipart start */
	fprintf(sendmail, "--" MULTIPART "\n");
	fprintf(sendmail, "Content-Type: application/octet-stream\n");
	fprintf(sendmail, "Content-Disposition: attachment; "
					"filename=\"scc%s.pgp\"\n", serialname());
	fprintf(sendmail, "\n");
}

static void
do_mail(hash_t *cgihash, rope_t *trope)
{
	char	*pgpid;			/* PGP User ID from form */
	FILE	*sendmail;		/* We all talk to sendmail */
	FILE	*pgp;			/* Lex talks to PGP */
	int		 outfds[2];		/* For a pipe to PGP */
	int		 infds[2];		/* For a pipe from PGP */

	sendmail = popen(SENDMAIL, "w");
	if (!sendmail) {
		log("cgisys", PRI_CRIT, "Could not popen sendmail (%s): %s",
						SENDMAIL, strerror(errno));
		exit(0);
	}

	mail_header(cgihash, sendmail);

	pgpid = hash_get(cgihash, "PGPuserid");
	if (pgpid && strlen(pgpid)) {
		pgp_header(cgihash, sendmail);
		pipe(outfds);
		pipe(infds);
		
		if (fork() == 0) {
			dup2(outfds[0], 0);		/* Pipe to stdin */
			close(outfds[1]);		/* Other end */
			dup2(infds[1], 1);		/* Pipe from stdout */
			close(infds[0]);		/* Other end */

			if (chdir(PGPDIR) != 0) {
				log("cgisys", PRI_CRIT, "Could not chdir to PGPDIR "
						PGPDIR ": %s", strerror(errno));
				exit(0);
			}

			execl(PGP, "-feat", pgpid, NULL);
			log("cgisys", PRI_CRIT, "Could not execl pgp "
					PGP "-feat %s: %s", pgpid, strerror(errno));
			exit(0);
		} /* END OF FORK */

		close(outfds[0]);
		close(infds[1]);

		if (fork() == 0) {
			sendmail_bounce(infds[0], sendmail);
			exit(0);
		} /* END OF FORK */

		close(infds[0]);

		pgp = fdopen(outfds[1], "w");
		if (!pgp) {
			log("cgisys", PRI_CRIT, "Could not fdopen PGP input "
							"filedes: %s", strerror(errno));
			exit(0);
		}
		parser_parse(cgihash, trope, pgp);
	}
	else {
		fprintf(sendmail, "\n");
		parser_parse(cgihash, trope, sendmail);
	}
}

void
invoke(void)
{
	char	*hostname;
	char	*template;
	hash_t	*cgihash;
	rope_t	*trope;
	rope_t	*mrope;

	cgihash = cgi_hash_new();

	hostname = getenv(ENV_HOSTNAME);
	if (!hostname) {
		log("cgiconf", PRI_CRIT, "No " ENV_HOSTNAME " in environment");
		exit(0);
	}
	hostname[strcspn(hostname, "/")] = 0;

	template = hash_get(cgihash, "case");
	if (!template) {
		log("cgiconf", PRI_CRIT, "No 'case' in hash table");
		exit(0);
	}
	template[strcspn(template, "/")] = 0;

	trope = rope_new(DIR_DATA "/");
	rope_add(trope, hostname);
	rope_add(trope, "/emailer/templates/");
	rope_add(trope, template);
	rope_add(trope, ".tpt");

	/* Here we have a chance to provide a totally nonstandard
	 * path to the parser, which will otherwise use directory
	 * structure rules. */

	mrope = rope_new(DIR_DATA "/");
	rope_add(mrope, hostname);
	rope_add(mrope, "/emailer/templates/message.tpt");

	if (access(rope_chars(trope), R_OK) != 0) {
		log("cgiconf", PRI_CRIT, "Could not access template '%s': %s",
						rope_chars(trope), strerror(errno));
		exit(0);
	}

	if (access(rope_chars(mrope), R_OK) != 0) {
		log("cgiconf", PRI_CRIT, "Could not access template '%s': %s",
						rope_chars(mrope), strerror(errno));
	}

	config_run(cgihash, hostname, template, mrope);
	do_mail(cgihash, trope);
}
